from __future__ import annotations

from functools import lru_cache

import astropy.constants as const
import astropy.units as u
import numpy as np
from numpy.typing import NDArray

h = const.h.value
c = const.c.value
k_B = const.k_B.value
T_sun = 5778  # K

SPECIFIC_INTENSITY_UNITS = u.W / u.Hz / u.m**2 / u.sr


def get_blackbody_emission(
    freq: float | NDArray[np.floating], T: float | NDArray[np.floating]
) -> float | NDArray[np.floating]:
    """Returns the blackbody emission given a frequency.

    Parameters
    ----------
    freq
        Frequency [GHz].
    T
        Temperature of the blackbody [K].

    Returns
    -------
        Blackbody emission [W / m^2 Hz sr].
    """

    freq *= 1e9
    term1 = (2 * h * freq**3) / c**2
    term2 = np.expm1(((h * freq) / (k_B * T)))

    return term1 / term2


def get_dust_grain_temperature(
    R: float | NDArray[np.floating], T_0: float, delta: float
) -> float | NDArray[np.floating]:
    """Returns the dust grain temperature given a radial distance from the Sun.

    Parameters
    ----------
    R
        Radial distance from the sun in ecliptic heliocentric coordinates [AU / 1AU].
    T_0
        Temperature of dust grains located 1 AU from the Sun [K].
    delta
        Powerlaw index.

    Returns
    -------
        Dust grain temperature [K].
    """

    return T_0 * R**-delta


def get_scattering_angle(
    R_los: float | NDArray[np.floating],
    R_helio: NDArray[np.floating],
    X_los: NDArray[np.floating],
    X_helio: NDArray[np.floating],
) -> NDArray[np.floating]:
    """
    Returns the scattering angle between the Sun and a point along the
    line of sight.

    Parameters
    ----------
    R_los
        Distance from observer to a point along the line of sight.
    R_helio
        Distance from the Sun to a point along the line of sight.
    X_los
        vector pointing from the observer to a point along the line of sight.
    X_helio
        Heliocentric position of point along the line of sight.

    Returns
    -------
        Scattering angle.
    """

    cos_theta = (X_los * X_helio).sum(axis=0) / (R_los * R_helio)

    return np.arccos(-cos_theta)


def get_phase_function(
    Theta: NDArray[np.floating], C: tuple[float, float, float]
) -> NDArray[np.floating]:
    """Returns the phase function.

    Parameters
    ----------
    Theta
        Scattering angle.
    coeffs
        Phase function parameters.

    Returns
    -------
        The Phase funciton.
    """

    phase_normalization = _get_phase_normalization(C)

    return phase_normalization * (C[0] + C[1] * Theta + np.exp(C[2] * Theta))


@lru_cache
def _get_phase_normalization(C: tuple[float, float, float]) -> float:
    """Returns the analyitcal integral for the phase normalization factor N."""

    int_term1 = 2 * np.pi
    int_term2 = 2 * C[0]
    int_term3 = np.pi * C[1]
    int_term4 = (np.exp(C[2] * np.pi) + 1) / (C[2] ** 2 + 1)

    return 1 / (int_term1 * (int_term2 + int_term3 + int_term4))



from . import components, source_parameters
from ._model import model_registry

model_registry.register_model(
    name="dirbe",
    comps=components.DIRBE,
    spectrum=source_parameters.SPECTRUM_DIRBE,
    emissivities=source_parameters.EMISSIVITY_DIRBE,
    albedos=source_parameters.ALBEDO_DIRBE,
    solar_irradiance_model="dirbe",
    phase_coefficients=source_parameters.PHASE_FUNCTION_DIRBE,
)

model_registry.register_model(
    name="planck13",
    comps=components.DIRBE,
    spectrum=source_parameters.SPECTRUM_PLANCK,
    emissivities=source_parameters.EMISSIVITY_PLANCK_13,
)

model_registry.register_model(
    name="planck15",
    comps=components.PLANCK,
    spectrum=source_parameters.SPECTRUM_PLANCK,
    emissivities=source_parameters.EMISSIVITY_PLANCK_15,
)

model_registry.register_model(
    name="planck18",
    comps=components.PLANCK,
    spectrum=source_parameters.SPECTRUM_PLANCK,
    emissivities=source_parameters.EMISSIVITY_PLANCK_18,
)

model_registry.register_model(
    name="odegard",
    comps=components.PLANCK,
    spectrum=source_parameters.SPECTRUM_PLANCK,
    emissivities=source_parameters.EMISSIVITY_ODEGARD,
)



from __future__ import annotations

from dataclasses import dataclass, field

import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

from ._component import Component
from ._component_label import ComponentLabel

from._solar_irradiance_model import SolarIrradianceModel
from .solar_irradiance_models import solar_irradiance_model_registry
from .source_parameters import DELTA_DIRBE, T_0_DIRBE


@dataclass
class Model:
    """Interplanetary Dust model.

    This dataclass acts as a container for the various parameters in a
    Interplanetary Dust model and provides a method to interpolate / extrapolate
    the spectral model parameters to other frequencies or wavelengths.
    """

    name: str
    comps: dict[ComponentLabel, Component]
    spectrum: u.Quantity[u.Hz] | u.Quantity[u.m]
    emissivities: dict[ComponentLabel, tuple[float, ...]]
    albedos: dict[ComponentLabel, tuple[float, ...]] | None = None
    solar_irradiance_model: SolarIrradianceModel | None = None
    phase_coefficients: list[tuple[float, ...]] | None = None
    T_0: float = T_0_DIRBE
    delta: float = DELTA_DIRBE

    @property
    def n_comps(self) -> int:
        return len(self.comps)

    def interpolate_source_parameters(
        self, comp_label: ComponentLabel, freq: u.Quantity[u.GHz] | u.Quantity[u.m]
    ) -> tuple[float, float, tuple[float, float, float]]:
        """
        Returns interpolated/extrapolated source parameters for a component
        given a frequency.
        """

        freq = freq.to(self.spectrum.unit, equivalencies=u.spectral())
        emissivity_interpolator = interp1d(
            x=self.spectrum,
            y=self.emissivities[comp_label],
            fill_value="extrapolate",
        )
        emissivity = emissivity_interpolator(freq)

        if self.albedos is not None:
            albedo_interpolator = interp1d(
                x=self.spectrum,
                y=self.albedos[comp_label],
                fill_value="extrapolate",
            )
            albedo = albedo_interpolator(freq)
        else:
            albedo = 0.0

        if self.phase_coefficients is not None:
            phase_coefficient_interpolator = interp1d(
                x=self.spectrum,
                y=np.asarray(self.phase_coefficients),
                fill_value="extrapolate",
            )
            phase_coefficient = phase_coefficient_interpolator(freq)
        else:
            phase_coefficient = [0.0 for _ in range(3)]

        return emissivity, albedo, tuple(phase_coefficient)

    def __repr__(self) -> str:
        comp_names = [comp_label.value for comp_label in self.comps]
        repr = f"{type(self).__name__}( \n"
        repr += f"   name: {self.name!r},\n"
        repr += "   components: (\n"
        for name in comp_names:
            repr += f"      {name!r},\n"
        repr += "   ),\n"
        repr += "   thermal: True,\n"
        repr += f"   scattering: {self.albedos is not None},\n"
        repr += ")"

        return repr


@dataclass
class ModelRegistry:
    """Container for registered models."""

    _registry: dict[str, Model] = field(init=False, repr=False, default_factory=dict)

    @property
    def models(self) -> list[str]:
        """Returns a list of registered model names."""

        return list(self._registry.keys())

    def register_model(
        self,
        name: str,
        comps: dict[ComponentLabel, Component],
        spectrum: u.Quantity[u.Hz] | u.Quantity[u.m],
        emissivities: dict[ComponentLabel, tuple[float, ...]],
        albedos: dict[ComponentLabel, tuple[float, ...]] | None = None,
        solar_irradiance_model: str | None = None,
        phase_coefficients: list[tuple[float, ...]] | None = None,
        T_0: float = T_0_DIRBE,
        delta: float = DELTA_DIRBE,
    ) -> None:
        """Registers a new model.

        For an example of how to register a custom model, see the documentation
        on https://zodipy.readthedocs.io/en/latest/ (coming soon).

        Parameters
        ----------
        name
            String representing the name of the model. This is the name that is
            used for the 'model' argument when initializing `Zodipy`.
        comps
            Dict mapping `CompLabel`s to `Component` classes.
        spectrum
            The spectrum (frequency or length units) corresponding to the
            frequencies used to estimate the spectral parameters.
        emissivities
            Emissivity factor fits for each IPD component at the frequencies
            corresponding to 'spectrum'.
        albedos
            Albedo factor fits for each IPD component at the frequencies
            corresponding to 'spectrum'.
        solar_irradiance_model
            Model for the solar irradiance.
        phase_coefficients
            Coefficient fits for the phase function at the frequencies
            corresponding to 'spectrum'.
        T_0
            Dust grain temperature at 1 AU. Defaults to the DIRBE model value.
        delta
            Dust grain temperature powerlaw parameter describing how the
            temperature falls with radial distance from the Sun. Defaults to the
            DIRBE model value.
        """

        if (name := name.lower()) in self._registry:
            raise ValueError(f"a model by the name {name!s} is already registered.")

        if solar_irradiance_model is not None:
            solar_model = solar_irradiance_model_registry.get_model(
                solar_irradiance_model
            )
        else:
            solar_model = solar_irradiance_model

        if albedos is not None and solar_irradiance_model is None:
            raise ValueError(
                "albedos are provided but no solar irradiance model is provided."
            )
        self._registry[name] = Model(
            name=name,
            comps=comps,
            spectrum=spectrum,
            emissivities=emissivities,
            T_0=T_0,
            delta=delta,
            albedos=albedos,
            solar_irradiance_model=solar_model,
            phase_coefficients=phase_coefficients,
        )

    def get_model(self, name: str) -> Model:
        """Returns a registered model given a name."""

        if (name := name.lower()) not in self._registry:
            raise ValueError(
                f"{name!r} is not a registered Interplanetary Dust model. "
                f"Avaliable models are: {', '.join(self._registry)}."
            )

        return self._registry[name]


model_registry = ModelRegistry()



from . import components, source_parameters
from ._component_label import ComponentLabel
from ._contour import tabulate_density
from .models import model_registry
from .solar_irradiance_models import solar_irradiance_model_registry
from .zodipy import Zodipy

__all__ = (
    "Zodipy",
    "model_registry",
    "solar_irradiance_model_registry",
    "components",
    "source_parameters",
    "tabulate_density",
    "ComponentLabel",
)



from __future__ import annotations

from functools import partial
from typing import Literal, Sequence, Union

import astropy.units as u
import healpy as hp
import numpy as np
import quadpy
from astropy.coordinates import solar_system_ephemeris
from astropy.time import Time
from numpy.typing import NDArray

from ._component import Component
from ._decorators import validate_ang, validate_freq, validate_pixels
from ._ephemeris import get_obs_and_earth_positions
from ._line_of_sight import get_line_of_sight_endpoints
from ._source_functions import (
    SPECIFIC_INTENSITY_UNITS,
    get_blackbody_emission,
    get_dust_grain_temperature,
    get_phase_function,
    get_scattering_angle,
)
from ._unit_vectors import get_unit_vectors_from_ang, get_unit_vectors_from_pixels
from .models import model_registry

DISTANCE_TO_JUPITER = u.Quantity(5.2, u.AU)
DEFAULT_SOLAR_CUTOFF = u.Quantity(5, u.deg)

HEALPixIndicies = Union[int, Sequence[int], NDArray[np.integer]]
SkyAngles = Union[u.Quantity[u.deg], u.Quantity[u.rad]]
FrequencyOrWavelength = Union[u.Quantity[u.Hz], u.Quantity[u.m]]


class Zodipy:
    """Interface for simulating zodiacal emission.

    Sets up the simulation configuration and provides methods for simulating the zodiacal
    emission that a solar system observer is predicted to see given an interplanetary dust
    model.

    Attributes:
        model (str): Name of the interplanetary dust model to use in the simulations.
            Defaults to DIRBE.
        ephemeris (str): Ephemeris used to compute the positions of the observer and the
            Earth. Defaults to 'de432s' which requires downloading (and caching) a ~10MB
            file. For more information on available ephemeridis, please visit
            https://docs.astropy.org/en/stable/coordinates/solarsystem.html
        extrapolate (bool): If True all spectral quantities in the selected model are
            linearly extrapolated to the requested frequency/wavelength. If False, an
            Exception will be raised on requested frequencies/wavelengths outside of the
            valid model range. Default is False.
        gauss_quad_order (int): Order of the Gaussian-Legendre quadrature used to evaluate
            the brightness integral. Default is 50 points.
        cutoff (u.Quantity[u.AU]): Radial distance from the Sun at which all line of sights
            are truncated. Defaults to 5.2 AU which is the distance to Jupiter.
        solar_cutoff (u.Quantity[u.deg]): Cutoff angle from the sun in degrees. The emission
            for all the pointing with angular distance between the sun smaller than
            `solar_cutoff` are set to `np.nan`. This is due to the model singularity of the
            diffuse cloud component at the heliocentric origin. Such a cutoff may be
            useful when using simulated pointing, but actual scanning strategies are
            unlikely to look directly at the sun. This feature is turned of by setting this
            argument to `None`. Defaults to 5 degrees.

    """

    def __init__(
        self,
        model: str = "dirbe",
        ephemeris: str = "de432s",
        extrapolate: bool = False,
        gauss_quad_order: int = 100,
        cutoff: u.Quantity[u.AU] = DISTANCE_TO_JUPITER,
        solar_cutoff: u.Quantity[u.deg] | None = DEFAULT_SOLAR_CUTOFF,
    ) -> None:

        self.model = model_registry.get_model(model)
        self.ephemeris = ephemeris
        self.extrapolate = extrapolate
        self.cutoff = cutoff
        self.solar_cutoff = solar_cutoff.to(u.rad) if solar_cutoff is not None else solar_cutoff
        self.integration_scheme = quadpy.c1.gauss_legendre(gauss_quad_order)

    @property
    def supported_observers(self) -> list[str]:
        """Returns a list of available observers given an ephemeris."""

        return list(solar_system_ephemeris.bodies) + ["semb-l2"]

    @validate_freq
    @validate_ang
    def get_emission_ang(
        self,
        freq: FrequencyOrWavelength,
        theta: SkyAngles,
        phi: SkyAngles,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        lonlat: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated zodiacal emission given angles on the sky.

        The pointing, for which to compute the emission, is specified in form of angles on
        the sky given by `theta` and `phi`.

        Args:
            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
            theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                the celestial sphere. Must be in the range [0, π] rad. Units must be either
                radians or degrees.
            phi: Angular longitude coordinate of a point, or a sequence of points, on the
                celestial sphere. Must be in the range [0, 2π] rad. Units must be either
                radians or degrees.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            lonlat: If True, input angles (`theta`, `phi`) are assumed to be longitude and
                latitude, otherwise, they are co-latitude and longitude.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        unique_angles, indicies = np.unique(
            np.asarray([theta, phi]), return_inverse=True, axis=1
        )
        unit_vectors = get_unit_vectors_from_ang(
            coord_in=coord_in,
            theta=unique_angles[0],
            phi=unique_angles[1],
            lonlat=lonlat,
        )

        emission = self._compute_emission(
            freq=freq,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=indicies,
        )

        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)

    @validate_freq
    @validate_pixels
    def get_emission_pix(
        self,
        freq: FrequencyOrWavelength,
        pixels: HEALPixIndicies,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated zodiacal emission given pixel numbers.

        The pixel numbers represent the pixel indicies on a HEALPix grid with resolution
        given by `nside`.

        Args:
            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
            pixels: HEALPix pixel indicies representing points on the celestial sphere.
            nside: HEALPix resolution parameter of the pixels and the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        unique_pixels, indicies = np.unique(pixels, return_inverse=True)
        unit_vectors = get_unit_vectors_from_pixels(
            coord_in=coord_in,
            pixels=unique_pixels,
            nside=nside,
        )

        emission = self._compute_emission(
            freq=freq,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=indicies,
            pixels=unique_pixels,
            nside=nside,
        )

        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)

    @validate_ang
    @validate_freq
    def get_binned_emission_ang(
        self,
        freq: FrequencyOrWavelength,
        theta: SkyAngles,
        phi: SkyAngles,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        lonlat: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated binned zodiacal emission given angles on the sky.

        The pointing, for which to compute the emission, is specified in form of angles on
        the sky given by `theta` and `phi`. The emission is binned to a HEALPix map with
        resolution given by `nside`.

        Args:
            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
            theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                the celestial sphere. Must be in the range [0, π] rad. Units must be either
                radians or degrees.
            phi: Angular longitude coordinate of a point, or a sequence of points, on the
                celestial sphere. Must be in the range [0, 2π] rad. Units must be either
                radians or degrees.
            nside: HEALPix resolution parameter of the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            lonlat: If True, input angles `theta`, `phi` are assumed to be longitude and
                latitude, otherwise, they are co-latitude and longitude.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        unique_angles, counts = np.unique(
            np.asarray([theta, phi]), return_counts=True, axis=1
        )
        unique_pixels = hp.ang2pix(nside, *unique_angles, lonlat=lonlat)
        unit_vectors = get_unit_vectors_from_ang(
            coord_in=coord_in,
            theta=unique_angles[0],
            phi=unique_angles[1],
            lonlat=lonlat,
        )

        emission = self._compute_emission(
            freq=freq,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=counts,
            binned=True,
            pixels=unique_pixels,
            nside=nside,
        )

        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)

    @validate_freq
    @validate_pixels
    def get_binned_emission_pix(
        self,
        freq: FrequencyOrWavelength,
        pixels: HEALPixIndicies,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Returns the simulated binned zodiacal Emission given pixel numbers.

        The pixel numbers represent the pixel indicies on a HEALPix grid with resolution
        given by `nside`. The emission is binned to a HEALPix map with resolution given by
        `nside`.

        Args:
            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
            pixels: HEALPix pixel indicies representing points on the celestial sphere.
            nside: HEALPix resolution parameter of the pixels and the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """

        unique_pixels, counts = np.unique(pixels, return_counts=True)
        unit_vectors = get_unit_vectors_from_pixels(
            coord_in=coord_in,
            pixels=unique_pixels,
            nside=nside,
        )

        emission = self._compute_emission(
            freq=freq,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=counts,
            binned=True,
            pixels=unique_pixels,
            nside=nside,
        )

        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)

    def _compute_emission(
        self,
        freq: u.Quantity[u.GHz],
        obs: str,
        obs_time: Time,
        unit_vectors: NDArray[np.floating],
        indicies: NDArray[np.integer],
        binned: bool = False,
        obs_pos: u.Quantity[u.AU] | None = None,
        pixels: NDArray[np.integer] | None = None,
        nside: int | None = None,
    ) -> NDArray[np.floating]:
        """Computes the component-wise zodiacal emission."""

        observer_position, earth_position = get_obs_and_earth_positions(
            obs=obs, obs_time=obs_time, obs_pos=obs_pos
        )

        if self.solar_cutoff is not None:
            # The observer position is aquired in geocentric coordinates before being
            # rotated to ecliptic coordinates which means that we can find the
            # heliocentric origin by simply taking the negative of the observer position.
            # We set all unit_vectors with a angular distance from the suns smaller than
            # the cutoff to np.nan which will propagate to np.nan/hp.UNSEEN in the emission.
            ang_dist = hp.rotator.angdist(-observer_position, unit_vectors)
            unit_vectors[:, ang_dist < self.solar_cutoff.value] = np.nan

        if self.model.solar_irradiance_model is not None:
            solar_irradiance = (
                self.model.solar_irradiance_model.interpolate_solar_irradiance(
                    freq=freq, albedos=self.model.albedos, extrapolate=self.extrapolate
                )
            )
        else:
            solar_irradiance = 0

        start, stop = get_line_of_sight_endpoints(
            cutoff=self.cutoff.value,
            obs_pos=observer_position,
            unit_vectors=unit_vectors,
        )

        emission = np.zeros(
            (self.model.n_comps, hp.nside2npix(nside) if binned else indicies.size)
        )

        # Preparing quantities for broadcasting
        stop_expanded = np.expand_dims(stop, axis=-1)
        observer_position = np.expand_dims(observer_position, axis=-1)
        earth_position = np.expand_dims(earth_position, axis=-1)
        unit_vectors = np.expand_dims(unit_vectors, axis=-1)

        # For each component, compute the integrated line of sight emission
        for idx, (label, comp) in enumerate(self.model.comps.items()):
            source_parameters = self.model.interpolate_source_parameters(label, freq)
            emissivity, albedo, phase_coefficients = source_parameters

            # Here we create a partial function that will be passed to the
            # integration scheme. The arrays are reshaped to (d, n, p) where d
            # is the geometrical dimensionality, n is the number of different
            # pointings, and p is the number of integration points of the
            # quadrature.
            emission_comp_integrand = partial(
                _compute_comp_emission_at_step,
                start=start,
                stop=stop_expanded,
                X_obs=observer_position,
                X_earth=earth_position,
                u_los=unit_vectors,
                comp=comp,
                freq=freq.value,
                T_0=self.model.T_0,
                delta=self.model.delta,
                emissivity=emissivity,
                albedo=albedo,
                phase_coefficients=phase_coefficients,
                solar_irradiance=solar_irradiance,
            )

            integrated_comp_emission = self.integration_scheme.integrate(
                emission_comp_integrand, [-1, 1]
            )
            # We convert the integral from [-1, 1] back to [start, stop].
            integrated_comp_emission *= 0.5 * (stop - start)

            if binned:
                emission[idx, pixels] = integrated_comp_emission
            else:
                emission[idx] = integrated_comp_emission[indicies]

        if binned:
            # We multiply the binned map by the number of hits.
            emission[:, pixels] *= indicies

        return emission

    # def _remove_masked_pointing

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(model={self.model.name!r}, "
            f"ephemeris={self.ephemeris!r}, "
            f"extrapolate={self.extrapolate!r})"
        )

    def __str__(self) -> str:
        return repr(self.model)


def _compute_comp_emission_at_step(
    r: float | NDArray[np.floating],
    *,
    start: float,
    stop: float | NDArray[np.floating],
    X_obs: NDArray[np.floating],
    X_earth: NDArray[np.floating],
    u_los: NDArray[np.floating],
    comp: Component,
    freq: float,
    T_0: float,
    delta: float,
    emissivity: float,
    albedo: float,
    phase_coefficients: tuple[float, float, float],
    solar_irradiance: float,
) -> NDArray[np.floating]:
    """Returns the zodiacal emission at a step along a line of sight."""

    # Convert the line of sight range from [-1, 1] to the true ecliptic positions
    R_los = ((stop - start) / 2) * r + (stop + start) / 2

    X_los = R_los * u_los
    X_helio = X_los + X_obs
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    density = comp.compute_density(X_helio, X_earth=X_earth)
    temperature = get_dust_grain_temperature(R_helio, T_0, delta)
    blackbody_emission = get_blackbody_emission(freq, temperature)

    emission = (1 - albedo) * (emissivity * blackbody_emission)

    if albedo > 0:
        solar_flux = solar_irradiance / R_helio**2
        scattering_angle = get_scattering_angle(R_los, R_helio, X_los, X_helio)
        phase_function = get_phase_function(scattering_angle, phase_coefficients)

        emission += albedo * solar_flux * phase_function

    return emission * density



from __future__ import annotations

from dataclasses import asdict, dataclass, field

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.utils.data import download_file
from scipy.interpolate import interp1d

from ._component_label import ComponentLabel
from ._source_functions import SPECIFIC_INTENSITY_UNITS


@dataclass(frozen=True)
class SolarIrradianceModelFromURL:
    """Model that is downloaded and cached first time it is requested."""

    url: str
    spectrum_unit: u.Unit
    irradiance_unit: u.Unit
    skip_rows: int


@dataclass(frozen=True)
class SolarIrradianceModel:
    """Solar irradiance spectrum given some model."""

    name: str
    spectrum: u.Quantity[u.Hz] | u.Quantity[u.m]
    irradiance: u.Quantity[SPECIFIC_INTENSITY_UNITS] | u.Quantity[u.MJy / u.sr]

    @classmethod
    def from_url(
        cls,
        name: str,
        url: str,
        spectrum_unit: u.Unit,
        irradiance_unit: u.Unit,
        skip_rows: int = 0,
    ) -> SolarIrradianceModel:
        """
        Initialize a Solar irradiance model from a table on the web. The model
        is downloaded and cached the first time it is requested.
        """

        solar_irradiance_table = download_file(url, cache=True)

        spectrum, irradiance = np.loadtxt(
            solar_irradiance_table, skiprows=skip_rows, unpack=True
        )
        spectrum *= spectrum_unit
        irradiance *= irradiance_unit

        try:
            irradiance = irradiance.to(SPECIFIC_INTENSITY_UNITS)
        except u.UnitConversionError:
            # The irradiance is stored in units of wavelength so we convert to
            # frequency by multiplying by lambda^2/c.
            irradiance = (spectrum**2 / const.c * irradiance).to(
                SPECIFIC_INTENSITY_UNITS
            )

        return SolarIrradianceModel(name, spectrum, irradiance)

    def interpolate_solar_irradiance(
        self,
        freq: u.Quantity[u.Hz] | u.Quantity[u.m],
        albedos: dict[ComponentLabel, tuple[float, ...]] | None,
        extrapolate: bool,
    ) -> float:
        """Returns the interpolated / extrapolated solar irradiance."""

        if albedos is None:
            return 0

        freq = freq.to(self.spectrum.unit, equivalencies=u.spectral())
        interpolator = interp1d(
            x=self.spectrum.value,
            y=self.irradiance.value,
            fill_value="extrapolate" if extrapolate else None,
        )
        try:
            solar_flux = interpolator(freq.value)
        except ValueError:
            raise ValueError(
                f"Solar flux model {self.name!r} is only valid in the "
                f"[{self.spectrum.min().value}, {self.spectrum.max().value}] "
                f"{self.spectrum.unit} range."
            )

        solar_flux *= self.irradiance.unit
        solar_flux_specific_intensity_units = solar_flux.to(
            SPECIFIC_INTENSITY_UNITS, equivalencies=u.spectral()
        )

        return solar_flux_specific_intensity_units.value


@dataclass
class SolarIrradianceModelRegistry:
    """Container for registered Solar irradiance models."""

    _registry: dict[str, SolarIrradianceModel | SolarIrradianceModelFromURL] = field(
        default_factory=dict
    )

    def register_model_from_url(
        self,
        name: str,
        url: str,
        spectrum_unit: u.Unit,
        irradiance_unit: u.Unit,
        skip_rows: int = 0,
    ) -> None:
        """Registers a model whos spectrum is tabulated on the web."""

        if (name := name.lower()) in self._registry:
            raise ValueError(f"a model by the name {name!s} is already registered.")

        self._registry[name] = SolarIrradianceModelFromURL(
            url, spectrum_unit, irradiance_unit, skip_rows
        )

    def register_model_from_table(
        self,
        name: str,
        spectrum: u.Quantity[u.Hz] | u.Quantity[u.m],
        irradiance: u.Quantity[SPECIFIC_INTENSITY_UNITS] | u.Quantity[u.MJy / u.sr],
    ) -> None:
        """Registers a model from a spectra."""

        if (name := name.lower()) in self._registry:
            raise ValueError(f"a model by the name {name!s} is already registered.")

        self._registry[name] = SolarIrradianceModel(name, spectrum, irradiance)

    def get_model(self, name: str) -> SolarIrradianceModel:
        """Returns a registered Solar irradiance model given a name."""

        try:
            model = self._registry[name]
        except KeyError:
            raise ValueError(
                f"{name!r} is not a registered Solar irradiance model. "
                f"Avaliable models are: {', '.join(self._registry)}."
            )

        if isinstance(model, SolarIrradianceModelFromURL):
            return SolarIrradianceModel.from_url(name, **asdict(model))

        return model


solar_irradiance_model_registry = SolarIrradianceModelRegistry()



from functools import wraps
from typing import TYPE_CHECKING, Sequence, Union

import astropy.units as u
import healpy as hp
import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from .zodipy import Zodipy


def validate_freq(function):
    """
    Decorator that validates the user inputed frequency or wavelength. This
    function also converts the 'freq' argument to units of GHz.
    """

    @wraps(function)
    @u.quantity_input
    def wrapper(
        self,
        freq: Union[u.Quantity[u.Hz], u.Quantity[u.m]],
        *args,
        **kwargs,
    ):
        # `u.quantity_input` doesnt work with TYPE_CHECKING string annotations.
        zodipy: "Zodipy" = self

        if not zodipy.extrapolate:
            freq = freq.to(zodipy.model.spectrum.unit, equivalencies=u.spectral())
            spectrum_min = zodipy.model.spectrum.min()
            spectrum_max = zodipy.model.spectrum.max()

            if not (spectrum_min <= freq <= spectrum_max):
                raise ValueError(
                    f"model {zodipy.model.name!r} is only valid in the [{spectrum_min.value},"
                    f" {spectrum_max.value}] {zodipy.model.spectrum.unit} range."
                )

        freq = freq.to(u.GHz, equivalencies=u.spectral())

        return function(self=self, freq=freq, *args, **kwargs)

    return wrapper


def validate_ang(function):
    """Decorator that validates the user inputed angles."""

    @wraps(function)
    @u.quantity_input
    def wrapper(
        *args,
        theta: Union[u.Quantity[u.deg], u.Quantity[u.rad]],
        phi: Union[u.Quantity[u.deg], u.Quantity[u.rad]],
        **kwargs,
    ):
        lonlat = kwargs.get("lonlat")
        theta = theta.to(u.deg) if lonlat is not None else theta.to(u.rad)
        phi = phi.to(u.deg) if lonlat is not None else phi.to(u.rad)

        if theta.isscalar:
            theta = u.Quantity([theta])
        if phi.isscalar:
            phi = u.Quantity([phi])

        return function(theta=theta, phi=phi, *args, **kwargs)

    return wrapper


def validate_pixels(function):
    """Decorator that validates the user inputed pixels."""

    @wraps(function)
    def wrapper(
        *args,
        pixels: Union[int, Sequence[int], NDArray[np.integer]],
        nside: int,
        **kwargs,
    ):

        if np.max(pixels) > hp.nside2npix(nside):
            raise ValueError("invalid pixel number given nside")

        if np.ndim(pixels) == 0:
            pixels = np.expand_dims(pixels, axis=0)

        return function(pixels=pixels, nside=nside, *args, **kwargs)

    return wrapper



import astropy.units as u

from ._solar_irradiance_model import solar_irradiance_model_registry
from .source_parameters import SOLAR_IRRADIANCE_DIRBE, SPECTRUM_DIRBE

# Temporary storage location for Solar irradiance tables
DOWNLOAD_URL = "http://tsih3.uio.no/www_cmb/metins/"

solar_irradiance_model_registry.register_model_from_table(
    name="dirbe", spectrum=SPECTRUM_DIRBE, irradiance=SOLAR_IRRADIANCE_DIRBE
)

solar_irradiance_model_registry.register_model_from_url(
    name="thuillier",
    url=DOWNLOAD_URL + "thuillier2004_flux.txt",
    spectrum_unit=u.Unit("nm"),
    irradiance_unit=u.Unit("mW/(m^2 nm sr)"),
    skip_rows=2,
)

solar_irradiance_model_registry.register_model_from_url(
    name="gueymard",
    url=DOWNLOAD_URL + "gueymard2003_flux.txt",
    spectrum_unit=u.Unit("nm"),
    irradiance_unit=u.Unit("W/(m^2 nm sr)"),
    skip_rows=9,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zodipy']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=5.0.1',
 'healpy>=1.15.0,<2.0.0',
 'jplephem>=2.17,<3.0',
 'numpy>=1.22.3,<2.0.0',
 'quadpy>=0.16.13,<0.17.0',
 'typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'zodipy',
    'version': '0.6.2',
    'description': 'Software for simulating zodiacal emission',
    'long_description': '\n<img src="docs/img/zodipy_logo.png" width="350">\n\n[![PyPI version](https://badge.fury.io/py/zodipy.svg)](https://badge.fury.io/py/zodipy)\n![Tests](https://github.com/MetinSa/zodipy/actions/workflows/tests.yml/badge.svg)\n[![astropy](http://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat)](http://www.astropy.org/)\n\n\n---\n\n\nZodiPy is a Python tool for simulating the Interplanetary Dust Emission that a Solar System observer sees, either in the form of timestreams or binned HEALPix maps.\n\n![plot](docs/img/zodipy_map.png)\n\n\n# Usage\nSee the [documentation](https://cosmoglobe.github.io/zodipy/) for a broader introduction to using ZodiPy.\n\n**Interplanetary Dust models:** select between built in models.\n```python\nfrom zodipy import Zodipy\n\nmodel = Zodipy(model="Planck18")\n```\n\n**Get emission from a point on the sky:** choose a frequency/wavelength, an observer, a time of observation, and angular coordinates (co-latitude, longitude).\n```python\nimport astropy.units as u\nfrom astropy.time import Time\n\nmodel.get_emission_ang(\n    25*u.micron,\n    theta=10*u.deg,\n    phi=40*u.deg,\n    obs="earth",\n    obs_time=Time.now(),\n)\n>> <Quantity [16.65684599] MJy / sr>\n```\n\n**Get emission from a sequence of angular coordinates:** `theta` and `phi` can be a sequence of angles that can represent some time-ordered pointing.\n```python\ntheta = [10.1, 10.5, 11.1, 11.5] * u.deg\nphi = [40.2, 39.9, 39.8, 41.3] * u.deg\n\nmodel.get_emission_ang(\n    25*u.micron,\n    theta=theta,\n    phi=phi,\n    obs="earth",\n    obs_time=Time.now(),\n    lonlat=True,\n)\n>> <Quantity [29.11106315, 29.33735654, 29.41248579, 28.30858417] MJy / sr>\n```\n\n\n**Get emission from pixel indices on a HEALPIX grid:** a sequence of pixel indicies along with an NSIDE parameter can be used.\n```python\nmodel.get_emission_pix(\n    25*u.micron,\n    pixels=[24654, 12937, 26135],\n    nside=128,\n    obs="earth",\n    obs_time=Time.now(),\n)\n>> <Quantity [17.77385144, 19.7889428 , 22.44797121] MJy / sr>\n```\n\n**Get binned emission component-wise:** the emission can be binned to a HEALPix map, and also returned component-wise.\n```python\nimport healpy as hp\nimport numpy as np\n\nnside = 128\n\nmodel.get_binned_emission_pix(\n    25*u.micron,\n    pixels=np.arange(hp.nside2npix(nside)),\n    nside=nside,\n    obs="earth",\n    obs_time=Time.now(),\n    return_comps=True\n).shape\n>> (6, 196608)\n```\n\n# Documentation\nA detailed introduction along with a tutorial of how to use ZodiPy will shortly be available in the [documentation](https://zodipy.readthedocs.io/en/latest/).\n# Installing\nZodiPy is available on PyPI and can be installed with ``pip install zodipy`` (Python >= 3.8 required).\n\n# Scientific paper\nSee [CITATION](https://github.com/Cosmoglobe/zodipy/blob/dev/CITATION.bib)\n\n- [Cosmoglobe: Simulating Zodiacal Emission with ZodiPy (San et al. 2022)](https://arxiv.org/abs/2205.12962)\n\n# Funding\nThis work has received funding from the European Union\'s Horizon 2020 research and innovation programme under grant agreements No 776282 (COMPET-4; BeyondPlanck), 772253 (ERC; bits2cosmology) and 819478 (ERC; Cosmoglobe).\n\n<table align="center">\n    <tr>\n        <td><img src="https://user-images.githubusercontent.com/28634670/170697040-d5ec2935-29d0-4847-8999-9bc4eaa59e56.jpeg" height="200"></td>\n        <td><img src="https://user-images.githubusercontent.com/28634670/170697140-b010aa69-9f9a-44c0-b702-8a05ec0b6d3e.jpeg" height="200"></td>\n    </tr>\n</table>',
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MetinSa/zodipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<=3.10',
}


setup(**setup_kwargs)



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

EPS = float(np.finfo(float).eps)


def get_line_of_sight_endpoints(
    cutoff: float, obs_pos: NDArray[np.floating], unit_vectors: NDArray[np.floating]
) -> tuple[float, NDArray[np.floating]]:
    """Returns the start and stop positions along the line of sights."""

    x, y, z = obs_pos.flatten()

    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arctan2(y, x)

    x_0 = r * np.cos(theta)
    y_0 = r * np.sin(theta)

    u_x, u_y, u_z = unit_vectors
    lon = np.arctan2(u_y, u_x)
    lat = np.arcsin(u_z)

    cos_lat = np.cos(lat)
    b = 2 * (x_0 * cos_lat * np.cos(lon) + y_0 * cos_lat * np.sin(lon))
    c = r**2 - cutoff**2
    q = -0.5 * b * (1 + np.sqrt(b**2 - 4 * c) / np.abs(b))

    return EPS, np.maximum(q, c / q)
